@@ -2,14 +2,21 @@
 
 from __future__ import annotations
 
+from functools import reduce
 from itertools import permutations
 
 import awkward as ak
 from awkward._backends.dispatch import backend_of_obj
-from awkward._backends.numpy import NumpyBackend
 from awkward._dispatch import high_level_function
 from awkward._do import mergeable
 from awkward._layout import HighLevelContext, ensure_same_backend, maybe_posaxis
+from awkward._namedaxis import (
+    NAMED_AXIS_KEY,
+    NamedAxesWithDims,
+    _get_named_axis,
+    _named_axis_to_positional_axis,
+    _unify_named_axis,
+)
 from awkward._nplikes.numpy_like import NumpyMetadata
 from awkward._nplikes.shape import unknown_length
 from awkward._parameters import type_parameters_equal
@@ -22,7 +29,6 @@ from awkward.types.numpytype import primitive_to_dtype
 __all__ = ("concatenate",)
 
 np = NumpyMetadata.instance()
-cpu = NumpyBackend.instance()
 
 
 @ak._connect.numpy.implements("concatenate")
@@ -94,7 +100,6 @@ def _merge_as_union(
 
 
 def _impl(arrays, axis, mergebool, highlevel, behavior, attrs):
-    axis = regularize_axis(axis)
     # Simple single-array, axis=0 fast-path
     if (
         # Is an array with a known backend
@@ -123,6 +128,15 @@ def _impl(arrays, axis, mergebool, highlevel, behavior, attrs):
             )
         )
 
+    # Handle named axis
+    merged_named_axis = reduce(_unify_named_axis, map(_get_named_axis, arrays))
+    # Step 1: normalize named axis to positional axis
+    axis = _named_axis_to_positional_axis(merged_named_axis, axis)
+    axis = regularize_axis(axis, none_allowed=False)
+    # Step 2: propagate named axis from input to output,
+    #   use strategy "unify" (see: awkward._namedaxis)
+    out_named_axis = merged_named_axis
+
     contents = [x for x in content_or_others if isinstance(x, ak.contents.Content)]
     if len(contents) == 0:
         raise ValueError("need at least one array to concatenate")
@@ -194,7 +208,8 @@ def _impl(arrays, axis, mergebool, highlevel, behavior, attrs):
                 nextinputs = []
                 for x in inputs:
                     if x.is_option and x.content.is_list:
-                        nextinputs.append(fill_none(x, [], axis=0, highlevel=False))
+                        empty = ak.to_backend([], backend)
+                        nextinputs.append(fill_none(x, empty, axis=0, highlevel=False))
                     else:
                         nextinputs.append(x)
                 inputs = nextinputs
@@ -247,7 +262,12 @@ def _impl(arrays, axis, mergebool, highlevel, behavior, attrs):
                     prototype[start : start + size] = tag
                     start += size
 
-                tags = ak.index.Index8(
+                if len(regulararrays) < 2**7:
+                    tags_cls = ak.index.Index8
+                else:
+                    tags_cls = ak.index.Index64
+
+                tags = tags_cls(
                     backend.index_nplike.reshape(
                         backend.index_nplike.broadcast_to(
                             prototype, (length, prototype.size)
@@ -266,10 +286,9 @@ def _impl(arrays, axis, mergebool, highlevel, behavior, attrs):
                 return (ak.contents.RegularArray(inner, prototype.size),)
 
             elif all(
-                isinstance(x, ak.contents.Content)
-                and x.is_list
-                or (isinstance(x, ak.contents.NumpyArray) and x.data.ndim > 1)
-                or not isinstance(x, ak.contents.Content)
+                (isinstance(x, ak.contents.Content) and x.is_list)  # Case 1
+                or (isinstance(x, ak.contents.NumpyArray) and x.data.ndim > 1)  # Case 2
+                or not isinstance(x, ak.contents.Content)  # Case 3: scalar value
                 for x in inputs
             ):
                 nextinputs = []
@@ -277,6 +296,8 @@ def _impl(arrays, axis, mergebool, highlevel, behavior, attrs):
                     if isinstance(x, ak.contents.Content):
                         nextinputs.append(x)
                     else:
+                        # Treat non-content inputs as scalars.
+                        # These become arrays of matching length.
                         nextinputs.append(
                             ak.contents.ListOffsetArray(
                                 ak.index.Index64(
@@ -303,7 +324,7 @@ def _impl(arrays, axis, mergebool, highlevel, behavior, attrs):
                 all_flatten = []
 
                 for x in nextinputs:
-                    o, f = x._offsets_and_flattened(1, 1)
+                    o, f = x._offsets_and_flattened(axis=1, depth=1)
                     c = o.data[1:] - o.data[:-1]
                     backend.index_nplike.add(counts, c, maybe_out=counts)
                     all_counts.append(c)
@@ -317,10 +338,15 @@ def _impl(arrays, axis, mergebool, highlevel, behavior, attrs):
 
                 offsets = ak.index.Index64(offsets, nplike=backend.index_nplike)
 
+                if len(nextinputs) < 2**7:
+                    tags_cls = ak.index.Index8
+                else:
+                    tags_cls = ak.index.Index64
                 tags, index = ak.contents.UnionArray.nested_tags_index(
                     offsets,
                     [ak.index.Index64(x) for x in all_counts],
                     backend=backend,
+                    tags_cls=tags_cls,
                 )
 
                 inner = ak.contents.UnionArray.simplified(
@@ -332,11 +358,35 @@ def _impl(arrays, axis, mergebool, highlevel, behavior, attrs):
             else:
                 return None
 
+        depth_context, lateral_context = NamedAxesWithDims.prepare_contexts(
+            list(arrays)
+        )
         out = ak._broadcasting.broadcast_and_apply(
-            content_or_others, action, allow_records=True, right_broadcast=False
+            content_or_others,
+            action,
+            depth_context=depth_context,
+            lateral_context=lateral_context,
+            allow_records=True,
+            right_broadcast=False,
         )[0]
+        # Unify named axes
+        out_named_axis = reduce(
+            _unify_named_axis, lateral_context[NAMED_AXIS_KEY].named_axis
+        )
 
-    return ctx.wrap(out, highlevel=highlevel)
+    wrapped_out = ctx.wrap(
+        out,
+        highlevel=highlevel,
+    )
+
+    # propagate named axis to output
+    return ak.operations.ak_with_named_axis._impl(
+        wrapped_out,
+        named_axis=out_named_axis,
+        highlevel=highlevel,
+        behavior=ctx.behavior,
+        attrs=ctx.attrs,
+    )
 
 
 def _form_has_type(form, type_):
